/*
  MLX90640 IR Temperature camera
  uses PJRC 320 x 240 TFT display
  and Teensy 3.6
  Based upon sample code by:
  Nathan Seidle
  SparkFun Electronics
  License: MIT. See license file for more information but you can
  basically do whatever you want with this code.

  This uses the driver written by Melexis found at:
  https://github.com/melexis/mlx90640-library

  I need a fast TFT display library- the default Adafruit one isn't nearly fast enough 
  TFT display didn't work properly with ILI9341_t3DMA  driver - refreshOnce and refresh.wait didn't work
  So, I used Kurt's ILI9341_t3n driver -this works fine @240 MHz and Optimize=Faster (but not using fastest nor LTO)
 at  https://github.com/KurtE/ILI9341_t3n
 This library needs SPIN library at 
 https://github.com/KurtE/SPIN
*/

#include <Wire.h>
#include "MLX90640_API.h"
#include "MLX90640_I2C_Driver.h"
#include "SPI.h"
#include "ILI9341_t3n.h"
#include <SD.h>

const byte MLX90640_address = 0x33; //Default 7-bit unshifted address of the MLX90640
#define TA_SHIFT 8 //Default shift for MLX90640 in open air
float mlx90640To[768];
float mlx90640Avg[768];  // used in weighted average filtering (in the time domain)
paramsMLX90640 mlx90640;
const byte scopeTrigger = 3; //Pin that goes high/low when calculations are complete
const byte RedLED = 4;
const byte GreenLED = 5;

// TFT display constants
#define TFT_DC 9
#define TFT_CS 10

// Dimensions used by SD card BMP file save routine
const uint16_t width = 320;      // landscape mode =Rotation(3)
const uint16_t height = 240;
const uint32_t filesize = 54 + 3 * width * height;
uint16_t pixelColorArray[3 * width];

ILI9341_t3n tft = ILI9341_t3n(TFT_CS, TFT_DC);  // using default SPI pins MOSI 11, SCK 13  MISO 12

float tempMin;
float tempMax;
float temps[768];
uint8_t r1;
uint8_t g1;
uint8_t b1;
uint16_t RGB1;
uint16_t RGB2;
uint16_t RGB3;
uint16_t RGB4;
int loopct = 0;

uint16_t x = 0;
uint16_t y = 0;
const uint16_t linesize = 3 * width;
uint8_t linebuf[linesize];

const int chipSelect = BUILTIN_SDCARD;  // for Teensy 3.6
File bmpFile;


#define DisplaySerial Serial
#define DEBUG

void setup()
{
  Serial.begin(115200); //Fast serial as possible
 // while (!Serial); //Wait for system monitor
 

  pinMode(scopeTrigger, OUTPUT);
  pinMode(RedLED, OUTPUT);
  pinMode(GreenLED, OUTPUT);
  digitalWrite(RedLED, LOW);    // bi-color LED is Common Cathode
  digitalWrite(GreenLED, LOW);
  pinMode(24, INPUT_PULLUP);
  if (!SD.begin(chipSelect)) {
	  Serial.println("Card failed, or not present");
	  digitalWrite(RedLED, HIGH);   // turn on RED LED
  }
  else {
	  Serial.println("uSD Card Initialized.");
	  digitalWrite(GreenLED, HIGH);
  }
  tft.begin();
  tft.setRotation(3);  // This is landscape mode with pin header to the LEFT

  Wire.begin();
  Wire.setClock(400000); //set I2C clock speed to 400kHz

  Serial.println("MLX90640 IR Camera");

  if (isConnected() == false)
  {
    Serial.println("MLX90640 not detected- Hang");
    while (1);
  }
  for (int i = 0; i < 768; i++) {     // zero out "last screen"  ARRAY used in weighted-averaging filter
	  mlx90640Avg[i] = 0;
  }

  //Get device parameters
  int status;
  uint16_t eeMLX90640[832];
  status = MLX90640_DumpEE(MLX90640_address, eeMLX90640);
  if (status != 0)
    Serial.println("Failed to load system parameters");

  status = MLX90640_ExtractParameters(eeMLX90640, &mlx90640);
  if (status != 0)
    Serial.println("Parameter extraction failed");

  //Once params are extracted, we can release eeMLX90640 array

  //Set refresh rate
  //MLX90640_SetRefreshRate(MLX90640_address, 0x03); //Set rate to 2Hz effective 
  MLX90640_SetRefreshRate(MLX90640_address, 0x04); //Set rate to 4Hz effective
  //MLX90640_SetRefreshRate(MLX90640_address, 0x05); //Set rate to 8Hz effective 

  //Once EEPROM has been read at 400kHz we can increase to 800 Khz
  Wire.setClock(1000000); //Teensy will now run I2C at 800kHz (because of clock division)
  FillScreen();
}

void loop()
{
	if (digitalRead(24) == LOW) {
		dumpScreenToSD();
	}
   for (byte x = 0 ; x < 2 ; x++)
  {
    uint16_t mlx90640Frame[834];
    int status = MLX90640_GetFrameData(MLX90640_address, mlx90640Frame);

 
    float vdd = MLX90640_GetVdd(mlx90640Frame, &mlx90640);
    float Ta = MLX90640_GetTa(mlx90640Frame, &mlx90640);
    float tr = Ta - TA_SHIFT; //Reflected temperature based on the sensor ambient temperature
    float emissivity = 0.95;
    MLX90640_CalculateTo(mlx90640Frame, &mlx90640, emissivity, tr, mlx90640To);
 
    }
   digitalWrite(scopeTrigger, HIGH);
  // do weighted average filtering 
  for (int i = 0; i < 768; i++) {
	  mlx90640Avg[i] += (mlx90640To[i] - mlx90640Avg[i]) * 0.25;
  }
  // get min and max temperature range
  tempMax = -999;
  tempMin = 999;
  for (int x = 0; x < 768; x++) {
	  if (mlx90640Avg[x] > tempMax) tempMax = mlx90640Avg[x];
	  if (mlx90640Avg[x] < tempMin) tempMin = mlx90640Avg[x];
  }
  for (int x = 0; x < 768; x++) {
	  temps[x] = map(mlx90640Avg[x], tempMin-0.01, tempMax+0.01, 256, 0);   // map tempMin to Blue and tempMax to Red
  }
  int x = 0;
  int y = 0;
  int i = 0;
  
  while (y < 205) {
	  float temp;
   // because of the way MLX90640 organizes its pixels (Right to left), we have to swap TFT x parameter  with 310-x
	  while (x < 320) {
		  HSVtoRGB(temps[i], 255, 255);
		  RGB1= tft.color565(r1, g1, b1);
		  HSVtoRGB(temps[i+1], 255, 255);
		  RGB2 = tft.color565(r1, g1, b1);
		  tft.fillRectHGradient(310 - x, y, 10, 5, RGB2, RGB1);
		  temp = (temps[i] +temps[i + 32]) / 2;
		  HSVtoRGB(temp, 255, 255);
		  RGB3 = tft.color565(r1, g1, b1);
		  temp = (temps[i + 1]+ temps[i + 33]) / 2;
		  HSVtoRGB(temp, 255, 255);
		  RGB4 = tft.color565(r1, g1, b1);
		  tft.fillRectHGradient(310 - x, y+5, 10, 4, RGB4, RGB3);
		  x = x + 10;
		  i++;
	  }
	  tft.fillRect(0 , y, 10, 5, RGB1);   // repeat the left-most mlx90640 pixel upper block (no further pixels to average)
	  tft.fillRect(0, y + 5, 10, 4, RGB3);  //  " " lower block
	  y = y + 9;
	  x = 0;
  }

  for (x = 0; x < 320; x += 10) {    // repeat the last horizontal line (no further lines to average)
	  float temp;
	  HSVtoRGB(temps[i], 255, 255);
	  RGB1 = tft.color565(r1, g1, b1);
	  HSVtoRGB(temps[i + 1], 255, 255);
	  RGB2 = tft.color565(r1, g1, b1);
	  tft.fillRectHGradient(310 - x, y, 10, 5, RGB2, RGB1);
	  i++;
  }
  digitalWrite(scopeTrigger, LOW);
  loopct++;
  if (loopct == 6) {   // only update temperature readings and gradient every 6 samples to eliminate text jitter
	  loopct = 0;
	  // draw a colour temperature gradient
	  int k = 256;
	  for (x = 0; x < 320; x += 10) {
		  HSVtoRGB(k, 255, 255);
		  tft.fillRect(x, 215, 10, 25, tft.color565(r1, g1, b1));
		  k = k - 8;
	  }
	  tft.setCursor(0, 220);
	  tft.setTextSize(2);
	  tft.setTextColor(ILI9341_WHITE);
	  tft.printf("%3.1fc", tempMin);
	  tft.setCursor(260, 220);
	  tft.setTextColor(ILI9341_BLACK);
	  tft.printf("%3.1fc", tempMax);
  }
}

//Returns true if the MLX90640 is detected on the I2C bus
boolean isConnected()
{
  Wire.beginTransmission((uint8_t)MLX90640_address);
  if (Wire.endTransmission() != 0)
    return (false); //Sensor did not ACK
  return (true);
}

void FillScreen() {
  tft.fillScreen(ILI9341_BLUE);
  delay(2000);
  yield();
 }


void HSVtoRGB(int hue, int sat, int val) {
	int base = 0;
	int r;
	int g;
	int b;

	switch (hue / 60) {
	case 0:
		r = val;
		g = (((val)*hue) / 60);
		b = base;
		break;
	case 1:
		r = (((val)*(60 - (hue % 60))) / 60);
		g = val;
		b = base;
		break;
	case 2:
		r = base;
		g = val;
		b = (((val)*(hue % 60)) / 60);
		break;
	case 3:
		r = base;
		g = (((val)*(60 - (hue % 60))) / 60);
		b = val;
		break;
	case 4:
		r = (((val)*(hue % 60)) / 60);
		g = base;
		b = val;
		break;
	case 5:
		r = val;
		g = base;
		b = (((val)*(60 - (hue % 60))) / 60);
		break;
	}
	r1 = r;
	g1 = g;
	b1= b;
}

void dumpScreenToSD(void) {

	uint8_t r, g, b;

	//-------------Local Declarations
	char filename[] = "IRCAM-xx.BMP";

	unsigned char bmpfileheader[14] =
	{ 'B','M', 0,0,0,0, 0,0, 0,0, 54,0,0,0 };


	struct xBITMAPINFOHEADER
	{
		uint32_t  biSize;           // size of this struct
		int32_t   biWidth;          // bmap width in pixels
		int32_t   biHeight;         // bmap height in pixels
		uint16_t  biPlanes;         // numplanes - always 1
		uint16_t  biBitCount;       // bits per pixel
		uint32_t  biCompression;    // compression flag
		uint32_t  biSizeImage;      // image size in bytes
		int32_t   biXPelsPerMeter;  // horz resolution
		int32_t   biYPelsPerMeter;  // vert resolution
		uint32_t  biClrUsed;        // 0 -> color table size
		uint32_t  biClrImportant;   // important color count
	};

	union ih_data {
		struct xBITMAPINFOHEADER ih;
		uint8_t ihData[sizeof(xBITMAPINFOHEADER)];
	} ihFrame;
	digitalWrite(RedLED, HIGH);
	digitalWrite(GreenLED, HIGH);    // LED= Yellow during SD file save process




	// Make a File Name & determine if it already exists 
	// IRCAMxx.BMP 
	// Open File
	for (uint8_t i = 0; i<100; i++) {
		filename[6] = i / 10 + '0';
		filename[7] = i % 10 + '0';
		if (!SD.exists(filename)) {
			// ony open a new file if it does not exist
			bmpFile = SD.open(filename, FILE_WRITE);
			break;
		}
	}

	Serial.println("SD.opened");

	if (!bmpFile) {
		Serial.print("Can't create file: ");
		Serial.println(filename);
		digitalWrite, (GreenLED, LOW);   // turn on RED LED
		return;
	}
	Serial.print("Opened new file: ");
	Serial.println(filename);
	//FORMAT FILE HEADER

	bmpfileheader[2] = (filesize);
	bmpfileheader[3] = (filesize >> 8);
	bmpfileheader[4] = (filesize >> 16);
	bmpfileheader[5] = (filesize >> 24);
	bmpFile.write(bmpfileheader, sizeof(bmpfileheader));

	//Populate Info HEADER
	ihFrame.ih.biSize = sizeof(xBITMAPINFOHEADER);
	ihFrame.ih.biWidth = width;
	ihFrame.ih.biHeight = height;
	ihFrame.ih.biPlanes = (uint16_t)1;
	ihFrame.ih.biBitCount = (uint16_t)24;
	ihFrame.ih.biCompression = (uint32_t)0;
	ihFrame.ih.biSizeImage = width * height;
	ihFrame.ih.biXPelsPerMeter = (uint32_t)0;
	ihFrame.ih.biYPelsPerMeter = (uint32_t)0;
	ihFrame.ih.biClrUsed = (uint32_t)0;
	ihFrame.ih.biClrImportant = (uint32_t)0;

	bmpFile.write( ihFrame.ihData, sizeof(xBITMAPINFOHEADER));
	
// Pixel info sent to file next
	for (int16_t i = height - 1; i >= 0; i--) {   // bitmap format is bottom line 1st upwards
		tft.readRect(0, i, width, 1, pixelColorArray);  // fetch a complete horizontal line
		for (int16_t j = 0; j < width; j++) {
			tft.color565toRGB(pixelColorArray[j], r, g, b);
			linebuf[j * 3] = b;   // in BMP format, blue is sent first
			linebuf[j * 3 + 1] = g;  // green
			linebuf[j * 3 + 2] = r;  // red
		}
		// write to SD card
		bmpFile.write(linebuf, linesize);
	}
	bmpFile.close();
	digitalWrite(RedLED, LOW);  // LED is now Green
	}

